<?php

class MM_WPFS_PhonyTransactionData {
    protected $customerEmail;

    public function __construct( $emailAddress ) {
        $this->customerEmail = $emailAddress;
    }

    /**
     * @return mixed
     */
    public function getCustomerEmail() {
        return $this->customerEmail;
    }
}

class MM_WPFS_PrivateAPI_v1 {
    protected static $instance;

    /**
     * @return MM_WPFS_PrivateAPI_v1
     */
    protected static function getInstance() {
        if ( self::$instance === null ) {
            self::$instance = new MM_WPFS_PrivateAPI_v1();
        }

        return self::$instance;
    }

    /**
     * @param $message
     * @return void
     */
    static public function log( $message ) {
        MM_WPFS_Utils::log( $message );
    }

    /**
     * @param $exception
     * @return void
     */
    static public function logException($exception, $object = null ) {
        MM_WPFS_Utils::logException( $exception, $object );
    }

    /**
     * @param $slug
     *
     * @return string
     */
    static public function getAdminUrlBySlug( $slug ) : string {
        return MM_WPFS_Admin_Menu::getAdminUrlBySlug( $slug );
    }

    /**
     * @param $slug
     * @param $params
     *
     * @return string
     */
    static public function getAdminUrlBySlugAndParams( $slug, $params = array() ) : string {
        return MM_WPFS_Admin_Menu::getAdminUrlBySlugAndParams( $slug, $params );
    }

    /**
     * @param $currency
     * @param $amount
     * @param $interval
     * @param $intervalCount
     *
     * @return string
     * @throws Exception
     */
    static public function formatPriceWithAdminSettings( $currency, $amount, $interval, $intervalCount ) {
        $formattedAmount = MM_WPFS_Currencies::formatAndEscapeByAdmin(
            $currency,
            $amount,
            true,
            true );

        return MM_WPFS_Localization::getPriceAndIntervalLabel( $interval, $intervalCount, $formattedAmount );
    }

    /**
     * @return string
     */
    public static function getWordpressTimezone() {
        return MM_WPFS_Utils::getWordpressTimezone();
    }

    /**
     * @param $timestamp
     *
     * @return string
     */
    public static function formatTimestampWithWordpressDateFormat( $timestamp ) {
        return MM_WPFS_Utils::formatTimestampWithWordpressDateFormat( $timestamp );
    }

    /**
     * @param $timestamp
     *
     * @return string
     */
    public static function formatTimestampWithWordpressDateTimeFormat( $timestamp ) {
        return MM_WPFS_Utils::formatTimestampWithWordpressDateTimeFormat( $timestamp );
    }

    /**
     * @param $liveMode int
     *
     * @return string
     */
    public static function getApiModeLabelFromInteger( $liveMode ) : string {
        return MM_WPFS_Admin::getApiModeLabelFromInteger( $liveMode );
    }

    /**
     * @param $assetName
     *
     * @return string
     */
    public static function getScriptAssetUrl( $assetName ) {
        return MM_WPFS_Assets::scripts( $assetName );
    }

    /**
     * @param $bindingResult
     * @return array
     */
    public static function generateReturnValueFromBindings( $bindingResult ) {
        return MM_WPFS_Utils::generateReturnValueFromBindings( $bindingResult );
    }

    /**
     * @param $elementId
     * @param $formHash
     * @param $index
     * @return string|void|null
     */
    public static function generateFormElementId( $elementId, $formHash, $index = null ) {
        return MM_WPFS_Utils::generateFormElementId( $elementId, $formHash, $index );
    }
}

class MM_WPFS_API_v1 {

    protected static $instance;

    /** @var MM_WPFS_Database */
    private $db = null;
    /** @var MM_WPFS_Stripe */
    protected $stripe;

    const STRIPE_API_MODE_CURRENT = 'current';
    const STRIPE_API_MODE_LIVE = 'live';
    const STRIPE_API_MODE_TEST = 'test';

    protected function __construct() {
        $this->db       = new MM_WPFS_Database();
        $this->stripe   = new MM_WPFS_Stripe( MM_WPFS::getStripeAuthenticationToken() );
    }

    /**
     * @return MM_WPFS_API_v1
     */
    protected static function getInstance() {
        if ( self::$instance === null ) {
            self::$instance = new MM_WPFS_API_v1();
        }

        return self::$instance;
    }

    /**
     * @return string
     */
    public static function getPluginVersion() {
        return MM_WPFS::VERSION;
    }

    /**
     * @return \StripeWPFS\StripeClient
     */
    public static function getStripeClient( $apiMode = self::STRIPE_API_MODE_CURRENT ) {
        switch( $apiMode ) {
            case self::STRIPE_API_MODE_CURRENT:
                return self::getInstance()->stripe->getStripeClient();

            case self::STRIPE_API_MODE_LIVE:
            case self::STRIPE_API_MODE_TEST:
                return MM_WPFS_Stripe::createStripeClient( MM_WPFS::getStripeAuthenticationTokenByMode( $apiMode === self::STRIPE_API_MODE_LIVE ) );

            default:
                throw new Exception( __CLASS__ . "." . __FUNCTION__ . "() - Invalid API mode: " . $apiMode );
        }
    }

    /**
     * @return array
     */
    public static function getStripeSubscriptionPlans() {
        return self::getInstance()->stripe->getSubscriptionPlans();
    }

    /**
     * @param $subscriptionId string
     * @return \StripeWPFS\Subscription
     * @throws \StripeWPFS\Exception\ApiErrorException
     */
    public static function getStripeSubscription($subscriptionId ) {
        return self::getInstance()->stripe->retrieveSubscription($subscriptionId );
    }

    /**
     * @param $subscriptionId
     * @param $params
     * @return \StripeWPFS\Subscription
     * @throws \StripeWPFS\Exception\ApiErrorException
     */
    public static function getStripeSubscriptionWithParams( $subscriptionId, $params ) {
        return self::getInstance()->stripe->retrieveSubscriptionWithParams( $subscriptionId, $params );
    }

    /**
     * @param $customerId
     * @return \StripeWPFS\Customer
     */
    public static function getStripeCustomer( $customerId ) {
        return self::getInstance()->stripe->retrieveCustomer( $customerId );
    }

    /**
     * @param $customerId
     * @param $params
     * @return \StripeWPFS\Customer
     */
    public static function getStripeCustomerWithParams( $customerId, $params ) {
        return self::getInstance()->stripe->retrieveCustomerWithParams( $customerId, $params );
    }

    /**
     * @param $params
     * @return \StripeWPFS\Collection
     * @throws \StripeWPFS\Exception\ApiErrorException
     */
    public static function getStripeCustomersWithParams( $params )  {
        return self::getInstance()->stripe->getCustomersWithParams( $params );
    }

    /**
     * @param $planId
     * @return \StripeWPFS\Price|null
     */
    public static function getStripePlan( $planId ) {
        return self::getInstance()->stripe->retrievePlan( $planId );
    }

    /**
     * @param $stripeCustomerId
     * @param $stripeSubscriptionId
     * @param $stripePlanId
     *
     * @throws \StripeWPFS\Exception\ApiErrorException
     */
    public static function changeSubscriptionPlan( $stripeCustomerId, $stripeSubscriptionId, $stripePlanId ) {
        $success = self::getInstance()->stripe->updateSubscriptionPlanAndQuantity( $stripeCustomerId, $stripeSubscriptionId, $stripePlanId );
        if ( $success ) {
            self::getInstance()->db->updateSubscriptionPlanByStripeSubscriptionId( $stripeSubscriptionId, $stripePlanId );
        }
    }

    /**
     * @param $stripeCustomerId
     * @param $stripeSubscriptionId
     * @return bool
     * @throws \StripeWPFS\Exception\ApiErrorException
     */
    public static function cancelSubscription( $stripeCustomerId, $stripeSubscriptionId ) {
        return self::getInstance()->stripe->cancelSubscription(
            $stripeCustomerId,
            $stripeSubscriptionId,
            MM_WPFS_Utils::getCancelSubscriptionsAtPeriodEnd()
        );
    }

    /**
     * @param $currency string
     * @return string
     */
    public static function getCurrencySymbolFor( $currency ) {
        return MM_WPFS_Currencies::getCurrencySymbolFor( $currency );
    }

    /**
     * @param $emailType string
     * @param $decoratedKeyValuePairs array
     */
    public static function sendEmailByTemplate( $recipientEmail, $templateType, $decoratedKeyValuePairs ) {
        if ( MM_WPFS_Utils::isDemoMode() ) {
            return;
        }

        $sender = new MM_WPFS_GenericEmailNotificationSender( new MM_WPFS_PhonyTransactionData( $recipientEmail ) );
        $sender->setTemplateType( $templateType );
        $sender->setDecoratedKeyValuePairs( $decoratedKeyValuePairs );
        $sender->sendEmail();
    }

    /**
     * @return bool
     */
    public static function isDemoMode() {
        return MM_WPFS_Utils::isDemoMode();
    }
}
